use std::collections::HashMap;
use std::time::Duration;

use dbus::{arg, Message};
use dbus::arg::RefArg;
use dbus::blocking::Connection;
use dbus::message::{MatchRule, SignalArgs};

use crate::background::{nitrogen_restore, process_image_url, reset_background_handler};
use crate::config::Config;
use crate::dbus_interface::main_dbus::DBusNameOwnerChanged;
use crate::dbus_interface::media_player::OrgMprisMediaPlayer2Player;
use crate::dbus_interface::statuses::PlayerStatus;
use crate::display::get_max_resolution;
use crate::result::{MBGError, MBGResult};

pub fn player_exit_handler(connection: &Connection) -> MBGResult<()> {
    connection.add_match(
        DBusNameOwnerChanged::match_rule(None, None),
        reset_background_handler,
    )?;
    Ok(())
}

pub fn player_state_watcher(connection: &Connection) -> MBGResult<()> {
    let mut property_changed_rule = MatchRule::new();
    property_changed_rule.path = Some("/org/mpris/MediaPlayer2".into());
    property_changed_rule.member = Some("PropertiesChanged".into());
    let (height, width) = get_max_resolution()?;
    connection.add_match(property_changed_rule, player_state_router(height, width))?;
    Ok(())
}

#[allow(clippy::implicit_hasher)]
pub fn process_metadata(
    metadata: HashMap<String, arg::Variant<Box<dyn arg::RefArg + 'static>>>,
    screen_height: i32,
    screen_width: i32,
) -> MBGResult<()> {
    let art_url = metadata.get("mpris:artUrl");
    if art_url.is_none() {
        return Err(MBGError::MetadataError(String::from(
            "mpris:artUrl not found in metadata",
        )));
    }
    let art_url = art_url.unwrap();
    let url = art_url.as_str();
    if url.is_none() {
        return Err(MBGError::MetadataError(String::from(
            "mpris:artUrl is not a string",
        )));
    }
    let url = url.unwrap().to_string();
    debug!("Found url: {}", url);
    process_image_url(url, screen_height, screen_width)?;
    Ok(())
}

pub fn player_state_router(screen_height: i32, screen_width: i32) -> impl FnMut((), &Connection, &Message) -> bool {
    move |_: (), connection: &Connection, message: &Message| {
        let current_status = PlayerStatus::from(message);
        debug!("Updated player state");
        match current_status {
            PlayerStatus::Playing => {
                if message.sender().is_none() {
                    return true;
                }
                let sender = message.sender().unwrap();
                let proxy =
                    connection.with_proxy(sender, "/org/mpris/MediaPlayer2", Duration::new(2, 0));
                let metadata = proxy.metadata();
                if let Err(d_bus_err) = metadata {
                    let my_err = MBGError::from(d_bus_err);
                    error!("{}", my_err);
                    return true;
                }
                let metadata = metadata.unwrap();
                let process_result = process_metadata(metadata, screen_height, screen_width);
                if let Err(error) = process_result {
                    error!("Found error while processing metadata.");
                    error!("{}", error);
                    return true;
                }
            }
            PlayerStatus::Paused => nitrogen_restore(),
            PlayerStatus::Stopped => nitrogen_restore(),
            PlayerStatus::Unknown => {}
        }
        true
    }
}

pub fn setup_connection() -> MBGResult<Connection> {
    let conn = dbus::blocking::Connection::new_session()?;
    player_exit_handler(&conn)?;
    player_state_watcher(&conn)?;
    Ok(conn)
}

pub fn start_listen() -> MBGResult<()> {
    let config = Config::get_props();
    debug!("Found config: \n{:#?}", config);
    let mut connection = setup_connection()?;
    if gtk::init().is_err() {
        return Err(MBGError::X11Error(String::from(
            "Failed to initialize GTK.",
        )));
    }

    loop {
        connection.process(Duration::from_millis(500)).unwrap();
    }
}
